import { useRef, useState, useEffect } from 'react';

const defaultReturn = {
    isValid: false,
    value: null,
};
function getRadioValue(options) {
    return Array.isArray(options)
        ? options.reduce((previous, { ref: { name, checked, value } }) => checked
            ? {
                isValid: true,
                value,
            }
            : previous, defaultReturn)
        : defaultReturn;
}

function getMultipleSelectValue(options) {
    return options.filter(({ selected }) => selected).map(({ value }) => value);
}

function isRadioInput(type) {
    return type === 'radio';
}

function getFieldValue(fields, { type, name, options, checked, value }) {
    if (isRadioInput(type)) {
        return getRadioValue(fields[name].options).value;
    }
    else if (type === 'select-multiple') {
        return getMultipleSelectValue(options);
    }
    else if (type === 'checkbox') {
        return checked;
    }
    return value;
}

function getFieldsValues(fields, filedNames) {
    return Object.values(fields).reduce((previous, { ref, ref: { name } }) => {
        const value = getFieldValue(fields, ref);
        if (typeof filedNames === 'string') {
            if (name === filedNames) {
                return value;
            }
            return previous;
        }
        const copy = Object.assign({}, (previous || {}));
        if (Array.isArray(filedNames)) {
            if (filedNames.includes(name)) {
                copy[name] = value;
            }
        }
        else {
            copy[name] = value;
        }
        return copy;
    }, undefined);
}

const DATE_INPUTS = ['date', 'time', 'month', 'datetime', 'datetime-local', 'week'];
const STRING_INPUTS = ['text', 'email', 'password', 'search', 'tel', 'url'];

function getValueAndMessage(item) {
    return {
        value: typeof item === 'object' && item.value ? item.value : item,
        message: typeof item === 'object' && item.message ? item.message : true,
    };
}

var validateField = ({ ref: { type, value, name, checked }, options, required, maxLength, minLength, min, max, pattern, validate }, fields) => {
    const copy = {};
    if (required &&
        ((type === 'checkbox' && !checked) ||
            (!isRadioInput(type) && type !== 'checkbox' && value === '') ||
            (isRadioInput(type) && !getRadioValue(fields[name].options).isValid))) {
        copy[name] = {
            required,
        };
    }
    if (value === '')
        return copy;
    // min and max section
    if (min || max) {
        let exceedMax;
        let exceedMin;
        const valueNumber = parseFloat(value);
        const { value: maxValue, message: maxMessage } = getValueAndMessage(max);
        const { value: minValue, message: minMessage } = getValueAndMessage(min);
        if (type === 'number') {
            exceedMax = maxValue && valueNumber > maxValue;
            exceedMin = minValue && valueNumber < minValue;
        }
        else if (DATE_INPUTS.includes(type)) {
            if (typeof maxValue === 'string')
                exceedMax = maxValue && new Date(value) > new Date(maxValue);
            if (typeof minValue === 'string')
                exceedMin = minValue && new Date(value) < new Date(minValue);
        }
        if (exceedMax || exceedMin) {
            copy[name] = Object.assign({}, copy[name], (exceedMax ? { max: maxMessage } : null), (exceedMin ? { min: minMessage } : null));
        }
    }
    if ((maxLength || minLength) && STRING_INPUTS.includes(type)) {
        const { value: maxLengthValue, message: maxLengthMessage } = getValueAndMessage(maxLength);
        const { value: minLengthValue, message: minLengthMessage } = getValueAndMessage(minLength);
        const exceedMax = maxLength && value.toString().length > maxLengthValue;
        const exceedMin = minLength && value.toString().length < minLengthValue;
        if (exceedMax || exceedMin) {
            copy[name] = Object.assign({}, copy[name], (exceedMax ? { maxLength: maxLengthMessage } : null), (exceedMin ? { minLength: minLengthMessage } : null));
        }
    }
    if (pattern) {
        const { value: patternValue, message: patternMessage } = getValueAndMessage(pattern);
        if (patternValue instanceof RegExp && !patternValue.test(value)) {
            copy[name] = Object.assign({}, copy[name], { pattern: patternMessage });
        }
    }
    if (validate) {
        const fieldValue = isRadioInput(type) ? getRadioValue(options).value : value;
        if (typeof validate === 'function') {
            const result = validate(fieldValue);
            if (typeof result !== 'boolean' || !result) {
                copy[name] = Object.assign({}, copy[name], { validate: result || true });
            }
        }
        else if (typeof validate === 'object') {
            const result = Object.entries(validate).reduce((previous, [key, validate]) => {
                const result = typeof validate === 'function' && validate(fieldValue);
                if (typeof result !== 'boolean' || !result) {
                    previous[key] = result || true;
                }
                return previous;
            }, {});
            if (Object.keys(result).length) {
                copy[name] = Object.assign({}, copy[name], { validate: result });
            }
        }
    }
    return copy;
};

var removeAllEventListeners = (ref, validateWithStateUpdate) => {
    ref.removeEventListener('input', validateWithStateUpdate);
    ref.removeEventListener('change', validateWithStateUpdate);
    ref.removeEventListener('blur', validateWithStateUpdate);
};

function findMissDomAndClean({ target: { ref, ref: { name, type }, mutationWatcher, options, }, fields, validateWithStateUpdate, forceDelete = false, }) {
    if (isRadioInput(type) && options) {
        options.forEach(({ ref }, index) => {
            if (!document.body.contains(ref) && fields[name] && fields[name].options && fields[name].options[index]) {
                removeAllEventListeners(fields[name].options[index], validateWithStateUpdate);
                fields[name].options[index].mutationWatcher.disconnect();
                fields[name].options.splice(index, 1);
            }
        });
        if (!fields[name].options.length) {
            delete fields[name];
            return fields;
        }
    }
    else if (ref && (!document.body.contains(ref) || forceDelete)) {
        removeAllEventListeners(ref, validateWithStateUpdate);
        if (mutationWatcher)
            mutationWatcher.disconnect();
        delete fields[name];
        return fields;
    }
    return fields;
}

function onDomRemove(element, onDetachCallback) {
    const observer = new MutationObserver(() => {
        function isDetached(el) {
            if (el.parentNode === window.document) {
                return false;
            }
            else if (el.parentNode === null) {
                return true;
            }
            return isDetached(el.parentNode);
        }
        if (isDetached(element)) {
            observer.disconnect();
            onDetachCallback();
        }
    });
    observer.observe(window.document, {
        childList: true,
        subtree: true,
    });
    return observer;
}

function attachEventListeners({ mode, allFields, watchFields, radioOptionIndex, ref, type, name, validateWithStateUpdate, }) {
    const field = allFields[name];
    const isOnChange = mode === 'onChange' || watchFields.current[ref.name];
    const isOnBlur = mode === 'onBlur';
    if (!field || (!isOnChange && !isOnBlur))
        return;
    const isRadio = isRadioInput(type);
    const event = isOnChange ? (isRadio ? 'change' : 'input') : 'blur';
    if (isRadio) {
        const options = field.options;
        const attachedEvents = options[radioOptionIndex].eventAttached;
        if (!options[radioOptionIndex])
            return;
        if (!attachedEvents || (attachedEvents && !attachedEvents.includes(event))) {
            options[radioOptionIndex].ref.addEventListener(event, validateWithStateUpdate);
            options[radioOptionIndex].eventAttached = [...attachedEvents || [], event];
        }
    }
    else {
        if (!field.eventAttached || (field.eventAttached && !field.eventAttached.includes(event))) {
            ref.addEventListener(event, validateWithStateUpdate);
            field.eventAttached = [...field.eventAttached || [], event];
        }
    }
}

function useForm({ mode } = { mode: 'onSubmit' }) {
    const fields = useRef({});
    const localErrorMessages = useRef({});
    const watchFields = useRef({});
    const [errors, updateErrorMessage] = useState({});
    function validateWithStateUpdate({ target: { name }, type }) {
        const ref = fields.current[name];
        const error = validateField(ref, fields.current);
        if (localErrorMessages.current[name] !== error[name] ||
            mode === 'onChange' ||
            (mode === 'onBlur' && type === 'blur') ||
            watchFields.current[name]) {
            const copy = Object.assign({}, localErrorMessages.current, error);
            if (!error[name])
                delete copy[name];
            localErrorMessages.current = copy;
            updateErrorMessage(copy);
        }
    }
    function removeReferenceAndEventListeners(data, forceDelete = false) {
        findMissDomAndClean({
            target: data,
            fields: fields.current,
            validateWithStateUpdate,
            forceDelete,
        });
    }
    function register(data) {
        if (!data || !data.ref)
            return;
        if (!data.ref.name) {
            console.warn('Oops missing the name for field:', data.ref);
            return;
        }
        let radioOptionIndex;
        const { ref, required, validate, ref: { name, type, value }, } = data;
        const allFields = fields.current;
        if (isRadioInput(type)) {
            if (!allFields[name]) {
                allFields[name] = { options: [], required, validate, ref: { type: 'radio', name } };
            }
            if (!allFields[name].validate && validate) {
                allFields[name].validate = validate;
            }
            const options = allFields[name].options || [];
            radioOptionIndex = options.findIndex(({ ref }) => value === ref.value);
            if (radioOptionIndex > -1) {
                options[radioOptionIndex] = Object.assign({}, options[radioOptionIndex], data);
            }
            else {
                options.push(Object.assign({}, data, { mutationWatcher: onDomRemove(ref, () => removeReferenceAndEventListeners(data, true)) }));
                radioOptionIndex = options.length - 1;
            }
        }
        else {
            allFields[name] = Object.assign({}, allFields[name], data);
            if (!allFields[name]) {
                allFields[name].mutationWatcher = onDomRemove(ref, () => removeReferenceAndEventListeners(data, true));
            }
        }
        attachEventListeners({ allFields, watchFields, ref, type, radioOptionIndex, name, mode, validateWithStateUpdate });
    }
    function watch(filedNames, defaultValue) {
        if (typeof filedNames === 'string') {
            if (!watchFields.current[filedNames])
                watchFields.current[filedNames] = true;
        }
        else if (Array.isArray(filedNames)) {
            filedNames.forEach(name => {
                if (!watchFields.current[name])
                    return;
                watchFields.current[name] = true;
            });
        }
        else {
            Object.values(fields.current).forEach(({ ref }) => {
                if (!ref)
                    return;
                if (!watchFields.current[ref.name])
                    return;
                watchFields.current[ref.name] = true;
            });
        }
        const result = getFieldsValues(fields.current, filedNames);
        return result === undefined ? defaultValue : result;
    }
    const handleSubmit = (callback) => e => {
        e.preventDefault();
        const allFields = fields.current;
        const { localErrors, values } = Object.values(allFields).reduce((previous, data) => {
            const { ref, ref: { name, type }, options, } = data;
            removeReferenceAndEventListeners(data);
            if (!fields.current[name])
                return previous;
            const fieldError = validateField(data, allFields);
            const hasError = fieldError && fieldError[name];
            if (!hasError) {
                previous.values[name] = getFieldValue(allFields, ref);
                return previous;
            }
            if (isRadioInput(type) && Array.isArray(options)) {
                options.forEach(option => {
                    if (option.eventAttached && option.eventAttached.includes('change'))
                        return;
                    option.ref.addEventListener('change', validateWithStateUpdate);
                    option.eventAttached = [...option.eventAttached, 'change'];
                });
            }
            else if (fields.current[name].eventAttached && !fields.current[name].eventAttached.includes('input')) {
                ref.addEventListener('input', validateWithStateUpdate);
                data.eventAttached = [...(data.eventAttached || []), 'input'];
            }
            previous.localErrors = Object.assign({}, (previous.localErrors || []), fieldError);
            return previous;
        }, {
            localErrors: {},
            values: {},
        });
        if (JSON.stringify(localErrorMessages.current) !== JSON.stringify(localErrors)) {
            updateErrorMessage(localErrors);
            localErrorMessages.current = localErrors;
        }
        if (!Object.values(localErrors).length)
            callback(values, e);
    };
    useEffect(() => () => {
        fields.current &&
            Object.values(fields.current).forEach(({ ref, options }) => isRadioInput(ref.type) && Array.isArray(options)
                ? options.forEach(({ ref }) => removeAllEventListeners(ref, validateWithStateUpdate))
                : removeAllEventListeners(ref, validateWithStateUpdate));
        fields.current = {};
        watchFields.current = {};
        localErrorMessages.current = {};
        updateErrorMessage({});
    }, [mode]);
    return {
        register,
        handleSubmit,
        errors,
        watch,
    };
}

export default useForm;
